    /*-------------------------------------------------------------------*/
    /* dhdsftpd.c                                                        */
    /*                                                                   */
    /* Radically cut-down version of FTP, to be started from inetd in    */
    /* place of the normal FTP, which allows anon access only, read      */
    /* access only, and access only to the single directory specified    */
    /* on the command line.                                              */
    /*                                                                   */
    /* This should compile on the default /bin/cc on a Sun-3 or Sun-4    */
    /* running SunOS 4.1 (preferably 4.1.1_U1 for Sun-3 and 4.1.3_U1     */
    /* for Sun-4).                                                       */
    /*                                                                   */
    /* Unfortunately this still has to run as root to bind port 20 to    */
    /* transfer data.                                                    */
    /*                                                                   */
    /* THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF EXNET SYSTEMS LTD. */
    /* The copyright notice(s) in this document do not evidence any      */
    /* actual or intended publication of such source code.               */
    /*                                                                   */
    /* Author: Damon Hart-Davis                                          */
    /* Start Date:  29/Apr/95                                            */
    /* Finish Date:                                                      */
    /*                                                                   */
    /* File 1 of 1                                                       */
    /*----------------------------------/--------------------------------*/

/*
NAME
    dhdsftpd---DHD's Safe FTP Daemon.

SYNOPSIS
    dhdsftpd [-t[est]] [-v[erbose]] [-l[og]] [-m[axload] load] directory

DESCRIPTION
    This should be run from inetd with a line something like:

        ftp stream tcp nowait root /usr/etc/dhdsftpd dhdsftpd -l /home/pub

    and will allow read-only anon access to the specified directory.

    Note that this has to run as root to bind to the ftp-data port, so
    is extra careful to check that files requested are plain files with
    global read permissions.

    This can be run `wait' rather than `nowait' to limit access to one
    daemon at a time.

    No directory-changing or upload commands are supported, and the
    daemon should run with as few privileges as possible.  This should
    be a very safe replacement for the real FTP.  Given that the daemon
    has to run as root, to guard against loopholes in the code this
    server should not be run on critcal machines, or machines with root
    access to critical resources, if at all possible.

    In logging mode all logins (with user ID, mail-ID and remote
    machine ID if possible) and session terminations are logged to
    syslog.  Serious errors (eg that result in early program
    termination) are always logged to syslog.

    In verbose mode more information, such as details of which files
    were requested and successfully transferred and how fast, are
    logged.  This *does not* affect the amount of information seen by
    the remote user so that a remote user cannot tell when their session
    is being logged.

    The `maxload' parameter determines the maximum system load at which
    we are prepared to operate.  When preparing to accept a connection
    the system load is tested.  If the load is higher than that
    specified as a maximum, the daemon exits.  If the load is higher
    than double the `maxload' value before any command, the daemon
    terminates.

    The `test' flag has the program write output to stdout rather than
    to file descriptor 0 so that it can be tested on the command-line
    taking on stdin/stdout.  Indetd normally leaves fd 0 set up for
    reading and writing.

    FTP commands understood are: USER PASS NOOP PORT TYPE RETR QUIT.

    The USER command argument must be `ftp' or `anonymous'; no other
    name will be accepted.

    The TYPE command can only specify A (ASCII) or I (Image), and no
    further parameters.

    The RETR command may only refer to a file in the current directory,
    ie may contain no `/'s in the name, and must be a plain file.

    Commands other than USER NOOP and QUIT will only be accepted when
    the user has logged in.  PASS will only be accepted immediately
    after USER.

BUGS
    Actually a feature.  This daemon only understands a very limited
    subset of FTP commands, and responds with only a fairly
    coarse-grained set of error/success codes.

    RFC959 says we must also accept MODE (stream only), STRU and STOR,
    but for the moment I don't.

    For the moment, the PORT command is NOT implemented, which will
    cause delays if multiple files are to be transferred.  However,
    many FTP clients (and Sun's in particular) can work round the
    absence of the PORT command moderately transparently.

    I don't actually look up MX records for the email address given as
    password to check that it does at least exist.

    I don't time out incoming idle connections, which I should.
 */

    static char *whatID = "@(#)V1.1.110 dhdsftpd.c 96/03/11, (c) ExNet Systems Ltd 1995--6";


    /*-------------------------------------------------------------------*/
    /* Includes                                                          */
    /*----------------------------------/--------------------------------*/
#define USE_RSTATD 1

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>                   /* For struct stat, etc. */
#include <sys/time.h>                   /* For struct timeval, etc. */
#include <sys/socket.h>                 /* For struct sockaddr, etc. */
#include <netinet/in.h>                 /* For struct sockaddr_in, etc. */
#include <fcntl.h>                      /* For open(). */
#include <syslog.h>                     /* For syslog(), openlog(), etc. */
#include <ctype.h>                      /* For isascii(), etc. */
#include <string.h>                     /* For strchr(), etc. */

#if USE_RSTATD
#include <rpcsvc/rstat.h>
#endif

    /* Missing header definition from stdio.h on some systems. */
#ifdef __STDC__
    extern FILE *fdopen(int fd, const char *type);
#else
    extern FILE *fdopen();
#endif

    /*-------------------------------------------------------------------*/
    /* Constants.                                                        */
    /*----------------------------------/--------------------------------*/
#define MAXIDLE (60 * 15)               /* Max idle time before the      */
                                        /* user's connection is dropped, */
                                        /* in seconds.                   */
                                        /*-------------------------------*/
#define FTP_DATA_PORT 20                /* Default FTP data port.        */
                                        /*-------------------------------*/
#define MAXparams 16                    /* Max number of words/parameters*/
                                        /* *including* command on input  */
                                        /* line.                         */
                                        /*-------------------------------*/
#define MAXline 781                     /* Max input line excluding      */
                                        /* trailing \r\n\0.              */
                                        /*-------------------------------*/
#define MAXcmd 4                        /* Max FTP command name in chars.*/
                                        /*-------------------------------*/
#define MAXpeername 1023                /* Max peer name in chars.       */
                                        /*-------------------------------*/

                                        /*-------------------------------*/
                                        /* State-machine states...       */
                                        /*-------------------------------*/
#define SMS_WaitingForUSER 0            /* Waiting for USER command.     */
#define SMS_WaitingForPASS 1            /* Waiting for PASS command.     */
#define SMS_LoggedIn       2            /* Logged in and ready.          */
                                        /*-------------------------------*/

                                        /*-------------------------------*/
                                        /* FTP response codes, generic.  */
                                        /*-------------------------------*/
#define FTPGRC_PPR      100             /* +ve preliminary reply.        */
#define FTPGRC_PCR      200             /* +ve completion reply.         */
#define FTPGRC_PIR      300             /* +ve intermediate reply.       */
#define FTPGRC_TNCR     400             /* Transient -ve completion reply*/
#define FTPGRC_PNCR     500             /* Permanent -ve completion reply*/
                                        /*-------------------------------*/
#define FTPGRC_SYNT     000             /* Syntax.                       */
#define FTPGRC_INFO     010             /* Information.                  */
#define FTPGRC_CON      020             /* Connections.                  */
#define FTPGRC_AUTH     030             /* Authentication and accounting.*/
#define FTPGRC_UNSP     040             /* Unspecified as yet.           */
#define FTPGRC_FS       050             /* File system.                  */
                                        /*-------------------------------*/

                                        /*-------------------------------*/
                                        /* FTP response codes.           */
                                        /*-------------------------------*/
#define FTPRC_FILEOK    150             /* File OK; opening data channel.*/
#define FTPRC_OK        200             /* Command OK.                   */
#define FTPRCsCMDUI_SUP 202             /* Command not impl, superfluous.*/
#define FTPRC_READY     220             /* Ready for new user/connection.*/
#define FTPRC_GOODBYE   221             /* Goodbye; closing ctl connctn. */
#define FTPRC_CLOSEDC   226             /* Closed data connection.       */
#define FTPRC_LOGGEDIN  230             /* User logged in, proceed.      */
#define FTPRC_NEEDPASS  331             /* Username OK, need password.   */
#define FTPRC_NOTAVAIL  421             /* Svc not avail, shutting down. */
#define FTPRC_CODCON    425             /* Can't open data connection.   */
#define FTPRC_SYNT      500             /* Syntax error, cmd unrecognised*/
#define FTPRC_SYNT_ARGS 501             /* Syntax error in args/params.  */
#define FTPRC_CMDUI     502             /* Command not implemented.      */
#define FTPRC_BADSEQ    503             /* Bad sequence of commands.     */
#define FTPRC_CMDUI_P   504             /* Command not imp for that param*/
#define FTPRC_NOTLOGGED 530             /* Not logged in.                */
#define FTPRC_NOTDONE   550             /* File unavailable.             */
                                        /*-------------------------------*/


#ifdef DEBUGGING_COMPILE
#define debugging_assert(ex) {if (!(ex)){(void)syslog(LOG_CRIT,"Assertion failed: file \"%s\", line %d\n", __FILE__, __LINE__);exit(1);}}
#else
#define debugging_assert(ex)
#endif

#ifdef __STDC__
#define CONST const
#define VOLATILE volatile
#else
#define CONST
#define VOLATILE
#endif


    /*-------------------------------------------------------------------*/
    /* Types.                                                            */
    /*----------------------------------/--------------------------------*/
#ifndef PRIVATE
#define PRIVATE static
#define FORWARD extern
    typedef int boolean;
#define true  1
#define false 0
#endif

#define FAIL -1
#define OK   0

    typedef unsigned char byte;

    /*-------------------------------------------------------------------*/
    /* Globals and externs.                                              */
    /*----------------------------------/--------------------------------*/
    PRIVATE char *progname;
    PRIVATE boolean verbose = false;
    PRIVATE boolean logging = false;


    /*-------------------------------------------------------------------*/
    /* system_load(m) returns the system load multiplied by 2 and        */
    /*     truncated to an integer.                                      */
    /*                                                                   */
    /* If `m' is true then the maximum of the short- and medium-term     */
    /* values is used, else just the short-term value is used.           */
    /*----------------------------------/--------------------------------*/
    PRIVATE int system_load(medium_load)
    CONST boolean medium_load;
        {
        struct statstime stat;

        int load;

#if USE_RSTATD
        if(rstat("localhost", &stat) != OK)
            {
            syslog(LOG_ERR, "failed to get stats from rstatd.");
            exit(1);
            }
#else
        ERROR---ALTERNATIVE CODE NOT YET WRITTEN.
#endif

        load = stat.avenrun[0];
        if(medium_load && (load < stat.avenrun[1])) { load = stat.avenrun[1]; }
        return(load >> 7);
        }


    /*-------------------------------------------------------------------*/
    /* usage() complains about argument format.                          */
    /*----------------------------------/--------------------------------*/
    PRIVATE void usage()
        {
        syslog(LOG_ERR,
"usage: %s [-t[est]] [-v[erbose]] [-l[og]] [-m[axload] load] directory.",
                progname);
        exit(1);
        }

    /*-------------------------------------------------------------------*/
    /* sanitise() converts a string entirely to printable ASCII, mapping */
    /*     all non-printable (eg < 32, >= 127) codes appropriately.      */
    /*                                                                   */
    /* The codes \t, \r and \n are left intact.                          */
    /*----------------------------------/--------------------------------*/
    PRIVATE void sanitise(s)
    register char *s;
        {
        debugging_assert(s);
        for( ; *s; ++s)
            {
            switch(*s) { case '\t': case '\n': case '\r': continue; }
            if((*s < 32) || (*s >= 127)) { *s = '?'; continue; }
            }
        }

    /*-------------------------------------------------------------------*/
    /* Goodbye exits without error.                                      */
    /*----------------------------------/--------------------------------*/
    PRIVATE void Goodbye(out)
    FILE *out;
        {
        debugging_assert(out);
        if(logging) { syslog(LOG_NOTICE, "goodbye."); }
        fprintf(out, "%d Goodbye.\r\n", FTPRC_GOODBYE);
        exit(0);
        }

    /*-------------------------------------------------------------------*/
    /* Close the control connection (abruptly?) with an error.           */
    /*----------------------------------/--------------------------------*/
    PRIVATE void Goodbye_ERROR(out, err)
    FILE *out;
    CONST char *err;
        {
        debugging_assert(out && err);
        if(logging) { syslog(LOG_NOTICE, "closing with error: %s.", err); }
        fprintf(out, "%d Error: %s.\r\n", FTPRC_NOTAVAIL, err);
        exit(1);
        }

    /*-------------------------------------------------------------------*/
    /* Report a syntax error in arguments.                               */
    /*----------------------------------/--------------------------------*/
    PRIVATE void SyntaxError(out, err)
    FILE *out;
    CONST char *err;
        {
        debugging_assert(out && err);
        if(verbose) { syslog(LOG_NOTICE, "syntax error in args: %s.", err); }
        fprintf(out, "%d Error in arguments: %s.\r\n", FTPRC_SYNT_ARGS, err);
        }

    /*-------------------------------------------------------------------*/
    /* Report wrong number of parameters.
    /*----------------------------------/--------------------------------*/
    PRIVATE void WrongNumberParams(out)
    FILE *out;
        { SyntaxError(out, "wrong number of parameters"); }

    /*-------------------------------------------------------------------*/
    /* Report generic `command OK'.                                      */
    /*----------------------------------/--------------------------------*/
    PRIVATE void CommandOK(out)
    FILE *out;
        {
        debugging_assert(out);
        if(verbose) { syslog(LOG_NOTICE, "command OK."); }
        fprintf(out, "%d OK.\r\n", FTPRC_OK);
        }

    /*-------------------------------------------------------------------*/
    /* Report generic `command OK' with comment.                         */
    /*----------------------------------/--------------------------------*/
    PRIVATE void CommandOK_comment(out, comment)
    FILE *out;
    CONST char *comment;
        {
        debugging_assert(out && comment);
        if(verbose) { syslog(LOG_NOTICE, "command OK; %s.", comment); }
        fprintf(out, "%d %s.\r\n", FTPRC_OK, comment);
        }


    /*-------------------------------------------------------------------*/
    /* read_command() reads the next command from the user, returning the*/
    /*     number of words found (including the command itself).         */
    /*                                                                   */
    /* The user's text is sanitised before being processed, to help      */
    /* avoid breaking things like syslog() later.                        */
    /*                                                                   */
    /* Words are separated by single spaces.                             */
    /*                                                                   */
    /* In addition, the first word is converted to uppercase and         */
    /* truncated if necessary to at most MAXcmd letters.                 */
    /*                                                                   */
    /* The args[] array must be able to hold at least MAXparams + 1      */
    /* string pointers.                                                  */
    /*                                                                   */
    /* The working buffer (buf) passed to read_command() must be at      */
    /* least MAXline + 3 bytes.                                          */
    /*----------------------------------/--------------------------------*/
    PRIVATE int read_command(buf, args)
    char buf[MAXline + 3];              /* Allow for trailing \r\n\0. */
    char *args[MAXparams + 1];
        {
        boolean got_nl = false;
        int i;
        int words;
        char *cp;
        int n = -1;
        boolean done;
        for(i = 0; i < MAXline + 2; ++i)
            {
            if((n = read(0, buf + i, 1)) != 1) /* EOF? */
                {
                if((i == 0) && (n == 0)) { args[0] = 0; return(0); }
                break;
                }
            if(buf[i] == '\n') { got_nl = true; break; }
            }
        buf[i] = '\0';
        debugging_assert(i < MAXline + 3);
        if(!got_nl && (n != 0))
            {
            syslog(LOG_ERR,
                   "overlength input line (%d chars max).",
                   MAXline);
            exit(1);
            }
        sanitise(buf);                  /* Get rid of any nasties. */

        done = false;
        for(args[words = 0] = cp = buf; !done && (words < MAXparams); )
            {
            char c = *cp++;
            debugging_assert(c != '\n');

            switch(c)
                {
                case '\r':              /* EOL; first of CRLF telnet EOL. */
                    cp[-1] = '\0';      /* Overwrite and drop through... */
                case '\0':              /* Missing '\r'; take to be EOL. */
                    { ++words; done = true; continue; }

                case ' ':               /* Separator. */
                    {
                    cp[-1] = '\0';      /* Turn into arg terminator. */
                    args[++words] = cp; /* Set up next argument. */
                    continue;
                    }
                }
            }
        args[words] = 0;                /* Terminate argument list. */

        if(words > 0)
            {
            if(strlen(args[0]) > MAXcmd)/* Truncate command name. */
                { args[0][MAXcmd] = '\0'; }
            for(cp = args[0]; *cp; ++cp)/* Convert to upper case. */
                { *cp = toupper(*cp); }
            }

        return(words);
        }


    /*-------------------------------------------------------------------*/
    /* retrieve_file() retrieves the file whose name is given by `name', */
    /*     in the mode indicated by `xfer_binary', to the user port      */
    /*     specified by `port'.  Progress, success and failure messages  */
    /*     will be written to the stream `out'.                          */
    /*                                                                   */
    /* The file requested will be checked to ensure it is not outside    */
    /* the current directory (ie contains no `/'s), that it is a plain   */
    /* file and has global read permissions.                             */
    /*                                                                   */
    /* Note that we allow the following for symlinks; one reason we      */
    /* don't do a chroot() at startup.                                   */
    /*                                                                   */
    /* For our end of the connection, we insist on the FTP_DATA_PORT     */
    /* but let the kernel choose the interface address to use.           */
    /*                                                                   */
    /* The connection made to the user port will be dropped when the     */
    /* transfer is complete.                                             */
    /*                                                                   */
    /* WE DON'T LOOK FOR `ABORT' COMMAND FROM THE USER AT THE MOMENT.    */
    /*----------------------------------/--------------------------------*/
    PRIVATE void retrieve_file(out, name, xfer_binary, portp)
    FILE *out;
    register CONST char *name;
    CONST boolean xfer_binary;
    CONST struct sockaddr_in *portp;
        {
        struct stat fstats;
        static CONST char *notavailable =
            { "%d File not available.\r\n" };
        static CONST char *noconnect =
            { "%d Can't open connection for transfer.\r\n" };
        static struct sockaddr_in our_port;
        int sd = -1;                    /* Socket for data channel. */
        FILE *f = 0;                    /* Descriptor for file. */
        FILE *fs = 0;                   /* Descriptor to write to socket. */
        int status = OK;
        int toggle;                     /* For setsockopt(). */

        debugging_assert(out && name && portp);
        debugging_assert((xfer_binary == true) || (xfer_binary == false));

        if(verbose)
            {
            syslog(LOG_NOTICE,
                   "RETR %s, mode %s...",
                   name,
                   xfer_binary ? "binary" : "ASCII");
            }

        if(strchr(name, '/'))
            {
            if(verbose) { syslog(LOG_ALERT, "non-local file requested."); }
            fprintf(out, notavailable, FTPRC_NOTDONE);
            return;
            }

        if(!(f = fopen(name, "r")) ||
           (fstat(fileno(f), &fstats) != OK) ||
           !S_ISREG(fstats.st_mode) ||
           ((fstats.st_mode & 0444) != 0444))
            {
            if(verbose)
                { syslog(LOG_NOTICE, "File absent, not plain or not a+r."); }
            fprintf(out, notavailable, FTPRC_NOTDONE);
            if(f) { fclose(f); }
            return;
            }

        if(verbose) { syslog(LOG_NOTICE, "file OK, opening data channel."); }
        fprintf(out,
                "%d Opening %s mode connection for %s (%d bytes).\r\n",
                FTPRC_FILEOK,
                xfer_binary ? "binary" : "ASCII",
                name,
                (int) (fstats.st_size));

        /*---------------------------------------------------------------*/
        /* Make connection to specified remote port.                     */
        /*                                                               */
        /* Set socket options to allow rebinding to the ftp-data port.   */
        /*------------------------------/--------------------------------*/
        if((status == OK) &&
           ((sd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0))
            {
            if(verbose) { syslog(LOG_ERR, "socket() failed."); }
            fprintf(out, noconnect, FTPRC_CODCON); status = FAIL;
            }
        bzero((char *)&our_port, sizeof(our_port));
        our_port.sin_family = AF_INET;
        our_port.sin_port = FTP_DATA_PORT;
        our_port.sin_addr.s_addr = INADDR_ANY;
        toggle = 1;
        if((status == OK) &&
           (setsockopt(sd, SOL_SOCKET,
                       SO_REUSEADDR, (char *)&toggle, sizeof(toggle)) != OK))
            {
            if(verbose) { syslog(LOG_ERR, "setsockopt() failed."); }
            fprintf(out, noconnect, FTPRC_CODCON); status = FAIL;
            }
        if((status == OK) &&
           (bind(sd, (struct sockaddr *)&our_port, sizeof(our_port)) != OK))
            {
            if(verbose) { syslog(LOG_ERR, "bind() failed."); }
            fprintf(out, noconnect, FTPRC_CODCON); status = FAIL;
            }
        if((status == OK) &&
           (connect(sd, (struct sockaddr *)portp, sizeof(*portp)) != OK))
            {
            if(verbose) { syslog(LOG_ERR, "connect() failed."); }
            fprintf(out, noconnect, FTPRC_CODCON); status = FAIL;
            }

        if((status == OK) &&
           !(fs = fdopen(sd, "w")))
            {
            if(verbose) { syslog(LOG_ERR, "fdopen() failed."); }
            fprintf(out, noconnect, FTPRC_CODCON); status = FAIL;
            }

        /*---------------------------------------------------------------*/
        /* Copy data from file to socket, expanding \n to \r\n if in     */
        /* ASCII mode.                                                   */
        /*                                                               */
        /* I should check for errors from putc(), and that I transfer    */
        /* right number of bytes in total, but I don't.                  */
        /*------------------------------/--------------------------------*/
        if(status == OK)
            {
            int ic;
            while((ic = getc(f)) != EOF)
                {
                if(!xfer_binary && (ic == '\n')) { putc('\r', fs); }
                putc(ic, fs);
                }
            }

        /*---------------------------------------------------------------*/
        /* Flush all buffers and close all open descriptors in an        */
        /* appropriate order.                                            */
        /*------------------------------/--------------------------------*/
        if(f)       { fclose(f); }
        if(fs)      { fflush(fs); fclose(fs); } /* Closes sd too? */
        if(sd >= 0) { close(sd); }
        if(status == OK)
            {
            if(verbose) { syslog(LOG_NOTICE, "Transfer complete."); }
            fprintf(out, "%d Transfer complete.\r\n", FTPRC_CLOSEDC);
            }
        else if(verbose) { syslog(LOG_NOTICE, "Transfer failed."); }
        }


    /*-------------------------------------------------------------------*/
    /* main_loop() is the main loop waiting for commands.                */
    /*                                                                   */
    /* When this exits a normal `goodbye' will be done.                  */
    /*                                                                   */
    /* We recheck system load just before acting on a user's command.    */
    /*----------------------------------/--------------------------------*/
    PRIVATE void main_loop(out, maxload, pa)
    FILE *out;
    CONST int maxload;
    CONST struct sockaddr_in *pa;       /* Peer address, AF_INET format. */
        {
        boolean xfer_binary = false;    /* Transfer ASCII by default.    */
        int state = SMS_WaitingForUSER; /* State-machine control...      */
        char buf[MAXline + 3];          /* Scratch for read_command().   */
        char *args[MAXparams + 1];      /* Pointers into buf[].          */
        struct sockaddr_in port;        /* User-DTP port for data xfer.  */
        int words;                      /* Number of words in input line.*/

        port = *pa;                     /* Set default port.             */

        for( ; ; )
            {
            int loadnow;
            /*-----------------------------------------------------------*/
            /* Read the next command from the user.                      */
            /*--------------------------/--------------------------------*/
            words = read_command(buf, args);
            debugging_assert(words >= 0);
            debugging_assert(words <= MAXparams);
            if(words == 0)
                {
                syslog(LOG_ERR, "unexpected EOF.");
                exit(1);
                }
            debugging_assert(args[0]);
            debugging_assert(!args[words]);
            if(verbose)
                {
                syslog(LOG_NOTICE,
                       "Got command %s with %d parameters.",
                       args[0],
                       words - 1);
                }

            /*-----------------------------------------------------------*/
            /* Recheck the system load...                                */
            /*--------------------------/--------------------------------*/
            if((loadnow = system_load(true)) > maxload * 4)
                {
                if(logging)
                    {
                    syslog(LOG_ERR,
                           "system load too high (%d, max %d).",
                           loadnow / 2,
                           maxload * 2);
                    }
                Goodbye_ERROR(out, "system load too high, sorry");
                }

            /*-----------------------------------------------------------*/
            /* Now process commands that can be accepted at any time.    */
            /*--------------------------/--------------------------------*/
            if(!strcmp(args[0], "QUIT")) /* Accept QUIT anytime. */
                { break; }

            if(!strcmp(args[0], "NOOP")) /* Accept NOOP anytime. */
                {
                if(words == 1) { CommandOK(out); }
                else { WrongNumberParams(out); }
                continue;
                }

            /*-----------------------------------------------------------*/
            /* Now check for commands that can only be accepted when the */
            /* user is logging in.                                       */
            /*--------------------------/--------------------------------*/
            if(!strcmp(args[0], "USER"))
                {
                if(state != SMS_WaitingForUSER)
                    {
                    if(verbose)
                        {
                        syslog(LOG_NOTICE,
                               "Can't do USER in state %d.\n",
                               state);
                        }
                    fprintf(out, "%d Can't change user.\r\n", FTPRC_NOTLOGGED);
                    continue;
                    }
                if(words != 2) { WrongNumberParams(out); continue; }
                debugging_assert(args[1]);
                if(strcmp(args[1], "ftp") && strcmp(args[1], "anonymous"))
                    {
                    syslog(LOG_ALERT,
                           "Tried to log in as `%s'.\n",
                           args[1]);
                    fprintf(out,
                            "%d Only anonymous access allowed.\r\n",
                            FTPRC_NOTLOGGED);
                    continue;
                    }
                state = SMS_WaitingForPASS;
                fprintf(out,
"%d Anonymous access OK, send your complete e-mail address as password.\r\n",
                        FTPRC_NEEDPASS);
                continue;
                }

            if(!strcmp(args[0], "PASS"))
                {
                if(state != SMS_WaitingForPASS)
                    {
                    if(verbose)
                        {
                        syslog(LOG_NOTICE,
                               "Can't do PASS in state %d.\n",
                               state);
                        }
                    fprintf(out, "%d Login with USER first.\r\n",FTPRC_BADSEQ);
                    continue;
                    }
                if(words != 2) { WrongNumberParams(out); continue; }
                debugging_assert(args[1]);
                if(!strchr(args[1], '@') || (strlen(args[1]) < 5))
                    {
                    static int counter;
                    static CONST char *ribbings[] =
                        {
                        "C'mon",
                        "Pull the other one",
                        "Be serious",
                        "Yeah, yeah",
                        "Hmm",
                        "Yeah, right",
                        "What!?",
                        "Uh huh?"
                        };

                    if(verbose)
                        {
                        syslog(LOG_NOTICE, "Email address without `@' tried.");
                        }
                    fprintf(out,
        "%d %s---that's not your full Internet email address is it?\r\n",
                            FTPRC_SYNT_ARGS,
                            ribbings[((unsigned)(++counter ^ getpid())) %
                                (sizeof(ribbings) / sizeof(ribbings[0]))]);
                    continue;
                    }
                if(logging || verbose)
                    { syslog(LOG_NOTICE, "Email address `%s'.", args[1]); }
                fprintf(out,
                        "%d Anon login OK, access restrictions apply.\r\n",
                        FTPRC_LOGGEDIN);
                state = SMS_LoggedIn;
                continue;
                }

            /*-----------------------------------------------------------*/
            /* Now check for commands that can only be accepted when the */
            /* user is logged in.                                        */
            /*--------------------------/--------------------------------*/
            if(state == SMS_LoggedIn)
                {
                if(!strcmp(args[0], "TYPE"))
                    {
                    char c;
                    if(words < 2) { WrongNumberParams(out); continue; }
                    if(words > 2)
                        {
                        fprintf(out,
                                "%d Only a single parameter implemented.\r\n",
                                FTPRC_CMDUI_P);
                        continue;
                        }
                    c = args[1][0];
                    if(c == '\0')
                        {
                        fprintf(out, "%d Missing type.\r\n", FTPRC_SYNT_ARGS);
                        continue;
                        }
                    switch(c)
                        {
                        case 'a':
                        case 'A':       /* ASCII */
                            {
                            xfer_binary = false;
                            CommandOK_comment(out, "ASCII transfer mode");
                            continue;
                            }

                        case 'i':
                        case 'I':       /* Image */
                            {
                            xfer_binary = true;
                            CommandOK_comment(out, "Binary transfer mode");
                            continue;
                            }

                        default:        /* Unsupported. */
                            { break; }
                        }
                    fprintf(out,
                            "%d Requested type not implemented.\r\n",
                            FTPRC_CMDUI_P);
                    continue;
                    }

                if(!strcmp(args[0], "RETR"))
                    {                   /* File transfer, here we come! */
                    if(words != 2) { WrongNumberParams(out); continue; }
                    debugging_assert(args[1]);
                    retrieve_file(out, args[1], xfer_binary, &port);
                    continue;
                    }

                }

            /*-----------------------------------------------------------*/
            /* Didn't recognise the command.                             */
            /*--------------------------/--------------------------------*/
            fprintf(out,
                    "%d `%s' command not understood.\r\n",
                    FTPRC_SYNT,
                    args[0]);
            }
        }


    /*-------------------------------------------------------------------*/
    /* main().                                                           */
    /*----------------------------------/--------------------------------*/
    int main(argc, argv)
    int argc;
    char **argv;
        {
        FILE *out;                      /* Output on socket.             */
        int maxload = 10;               /* Maximum acceptable load.      */
        char *dirname = 0;              /* Name of working directory.    */
        boolean test = false;           /* Write to stdout, not fd 0.    */

        int loadnow;                    /* Current load.                 */
        struct sockaddr peerad;         /* Peer address.                 */
        int peeradlen;

        /*---------------------------------------------------------------*/
        /* Initialise.                                                   */
        /*------------------------------/--------------------------------*/
        progname = *argv;
        openlog("dhdsftp", LOG_PID, LOG_DAEMON); /* `Lazy' open. */

        /*---------------------------------------------------------------*/
        /* Parse command-line args for optional flags.                   */
        /*------------------------------/--------------------------------*/
        for(++argv ; --argc && (**argv == '-'); ++argv)
            {
            char C;

            switch(C = (*argv)[1])
                {
                default:
                    {
                    syslog(LOG_ERR, "switch -%c not understood.", C);
                    usage();
                    }

                case 'l': { logging = true; break; }
                case 't': { test = true; break; }
                case 'v': { verbose = true; break; }

                case 'm':
                    {
                    int val;

                    if(--argc < 0)
                        {
                        syslog(LOG_ERR, "missing arg to -%c.", C);
                        usage();
                        }

                    val = atoi(*++argv);
                    maxload = val;
                    break;
                    }
                }
            }

        /*---------------------------------------------------------------*/
        /* Parse command-line args for manditory arguments.              */
        /*------------------------------/--------------------------------*/
        if(argc != 1)
            {
            syslog(LOG_ERR, "%d args instead of %d expected.", argc, 1);
            usage();
            }
        dirname = argv[0];

        /*---------------------------------------------------------------*/
        /* Do start-up logging messages as necessary.                    */
        /*------------------------------/--------------------------------*/
        if(logging || verbose)
            { syslog(LOG_NOTICE, "%s, dir %s.", whatID + 4, dirname); }
        if(logging) { syslog(LOG_NOTICE, "logging on."); }
        if(verbose) { syslog(LOG_NOTICE, "verbose on."); }

        /*---------------------------------------------------------------*/
        /* Get peer name always.                                         */
        /*------------------------------/--------------------------------*/
        peeradlen = sizeof(peerad);
        if((getpeername(0, &peerad, &peeradlen) != OK) ||
           (peeradlen != sizeof(peerad)) ||
           (peerad.sa_family != AF_INET))
            {
            if(!test) { syslog(LOG_ALERT, "failed to getpeername()."); }
            bzero((char *)&peerad, sizeof(peerad)); /* Clear to all zeros. */
            }
        else
            {
            if(verbose || logging)
                {
                CONST struct sockaddr_in *si = (struct sockaddr_in *)&peerad;
                CONST unsigned long saddr = ntohl(si->sin_addr.s_addr);
                syslog(LOG_NOTICE,
                       "connection from %d.%d.%d.%d port %d.",
                       (int)((saddr >> 24) & 0xff),
                       (int)((saddr >> 16) & 0xff),
                       (int)((saddr >>  8) & 0xff),
                       (int)((saddr      ) & 0xff),
                       (int)(si->sin_port));
                }
            }

        /*---------------------------------------------------------------*/
        /* Change to final directory.                                    */
        /*------------------------------/--------------------------------*/
        debugging_assert(dirname);
        if(chdir(dirname) != OK)
            {
            syslog(LOG_ALERT, "couldn't chdir(%s).", dirname);
            exit(1);
            }
        if(verbose) { syslog(LOG_NOTICE, "successful chdir(%s).", dirname); }

        /*---------------------------------------------------------------*/
        /* Connect output descriptor to socket...                        */
        /*------------------------------/--------------------------------*/
        if(test)
            {
            out = stdout;
            if(verbose) { syslog(LOG_NOTICE, "running in test mode."); }
            }
        else if(!(out = fdopen(0, "w")))
            {
            syslog(LOG_ERR, "can't open output connection to socket.");
            exit(1);
            }
        setlinebuf(out);                /* Use line buffering.           */

        /*---------------------------------------------------------------*/
        /* Check the system load...                                      */
        /*------------------------------/--------------------------------*/
        if((loadnow = system_load(true)) > maxload * 2)
            {
            if(logging)
                {
                syslog(LOG_ERR,
                       "system load too high at startup (%d, max %d).",
                       loadnow / 2,
                       maxload);
                }
            Goodbye_ERROR(out, "system load too high at startup, sorry");
            }

        /*---------------------------------------------------------------*/
        /* Say hello and enter the main loop.                            */
        /*------------------------------/--------------------------------*/
        fprintf(out, "%d anon FTP server %s.\r\n", FTPRC_READY, whatID + 4);
        main_loop(out, maxload, (struct sockaddr_in *)&peerad);

        /*---------------------------------------------------------------*/
        /* Say goodbye.                                                  */
        /*------------------------------/--------------------------------*/
        Goodbye(out);

        debugging_assert(false);        /* Shouldn't get here.           */
        return(0);
        }
